import os

import click

from .core import Combine
from .checks import CheckRunner
from .dev import Watcher, Server


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.command()
@click.option('--no-checks', is_flag=True, default=False)
@click.pass_context
def build(ctx, no_checks):
    content_paths = [
        os.path.abspath('content'),
        os.path.abspath(os.path.join('theme', 'content')),
    ]

    output_path = os.path.abspath('output')

    config_path = os.path.abspath('combine.yml')

    combine = Combine(
        config_path=config_path,
        content_paths=content_paths,
        output_path=output_path,
    )

    click.secho('Installing dependencies', fg='cyan')
    combine.install()

    click.secho('Building site', fg='cyan')
    combine.build()

    if not no_checks:
        runner = CheckRunner(combine)
        runner.run()
        if runner.succeeded:
            click.secho('All checks passed!', fg='green')
        else:
            click.secho('Checks failed.', fg='red')
            for check in runner.failed_checks:
                click.echo(check)
            exit(1)

    return combine


@cli.command()
@click.pass_context
def work(ctx):
    combine = ctx.invoke(build, no_checks=True)

    click.secho('Watching for file changes...', fg='green')

    server = Server(combine.output_path)
    watcher = Watcher('.', combine=combine)
    watcher.watch(server.serve)


# @cli.command()
# @click.pass_context
# def check(ctx):
#     combine = ctx.invoke(build, no_checks=True)  # should it build?
#     runner = CheckRunner(combine)
#     runner.run()
#     if runner.succeeded:
#         click.secho('All checks passed!', fg='green')
#     else:
#         click.secho('Checks failed.', fg='red')
#         for check in runner.failed_checks:
#             click.echo(check)
#         exit(1)


if __name__ == '__main__':
    cli()


# Want to make it as easy and friendly as possible, hard to screw up
# pretty print output html? or don't want to screw with js, etc.
# validate html
# check "output" in .gitignore (if git) -- use warning codes FW101
# asset pipeline (asset fingerprinting - hash)

# redirects
# → curl https://help.github.com/articles/creating-an-issue-template-for-your-repository/
# <!DOCTYPE html>
# <html>
# <head>
# <meta charset=utf-8>
# <title>Redirecting...</title>
# <link rel=canonical href="/articles/manually-creating-a-single-issue-template-for-your-repository/">
# <meta http-equiv=refresh content="0; url=/articles/manually-creating-a-single-issue-template-for-your-repository/">
# <h1>Redirecting...</h1>
# <a href="/articles/manually-creating-a-single-issue-template-for-your-repository/">Click here if you are not redirected.</a>
# <script>location='/articles/manually-creating-a-single-issue-template-for-your-repository/'</script>
# </body>
# </html>

# rel canonical?

# build in pre-commit hook helper -- you don't have a pre-commit hook, do you want one?
# then it runs check command before commit (not on every run if costly)

# no links to same page (empty links too)



import os
from subprocess import run

from .core import File
from .utils import create_parent_directory


class SassFile(File):
    # @classmethod
    # def class_pre_build_check(cls):
    #     run(['which', 'scss'], check=True)

    def render_to_output(self, output_path, *args, **kwargs):
        target_path = os.path.join(output_path, self.output_relative_path)
        create_parent_directory(target_path)

        target_root, scss_ext = os.path.splitext(target_path)
        css_target_path = target_root + '.css'

        run(['sass', self.path, css_target_path], check=True)



import os
import shutil
import subprocess

import jinja2

from .config import Config
from .files import file_class_for_path


class Combine:
    def __init__(self, config_path, content_paths, output_path):
        self.config_path = config_path
        self.content_paths = content_paths
        self.output_path = output_path
        self.load()

    def load(self):
        self.content_directories = [ContentDirectory(x) for x in self.content_paths if os.path.exists(x)]

        self.config = Config(self.config_path)

        choice_loaders = [jinja2.FileSystemLoader(x.path) for x in self.content_directories]

        self.jinja_environment = jinja2.Environment(
            loader=jinja2.ChoiceLoader(choice_loaders),
            autoescape=jinja2.select_autoescape(['html', 'xml']),
            undefined=jinja2.StrictUndefined,  # make sure variables exist
        )
        self.jinja_environment.globals = self.config.get_variables()

    def reload(self):
        """Reload the config and entire jinja environment"""
        self.load()

    def install(self):
        for cmd in self.config.get_commands('install'):
            subprocess.run(cmd, shell=True, check=True)

    def clean(self):
        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path)

    def pre_build_checks(self):
        for content_directory in self.content_directories:
            for file_class in content_directory.file_classes():
                file_class.class_pre_build_check()

            for file in content_directory.files:
                file.pre_build_check()

    def post_build_checks(self):
        for content_directory in self.content_directories:
            for file_class in content_directory.file_classes():
                file_class.class_post_build_check()

            for file in content_directory.files:
                file.post_build_check()

    def build(self, only_paths=None):
        self.pre_build_checks()

        if not only_paths:
            # completely wipe it
            self.clean()

        if not os.path.exists(self.output_path):
            os.mkdir(self.output_path)

        paths_rendered = []

        for content_directory in self.content_directories:
            for file in content_directory.files:
                if file.output_relative_path and file.output_relative_path not in paths_rendered:
                    if only_paths and file.path not in only_paths:
                        continue

                    file.render_to_output(
                        self.output_path,
                        jinja_environment=self.jinja_environment,
                    )
                    paths_rendered.append(file.output_relative_path)

        self.post_build_checks()

    def is_in_content_paths(self, path):
        for cp in self.content_paths:
            if os.path.commonpath([cp, path]) != os.getcwd():
                return True
        return False

    def is_in_output_path(self, path):
        return os.path.commonpath([self.output_path, path]) != os.getcwd()


class ContentDirectory:
    def __init__(self, path):
        self.path = path
        self.load_files()

    def load_files(self):
        self.files = []

        for root, dirs, files in os.walk(self.path):
            for file in files:
                file_path = os.path.join(root, file)
                self.files.append(
                    file_class_for_path(file_path)(file_path, self)
                )

    def file_classes(self):
        return set([x.__class__ for x in self.files])



from . import states
from .empty_build import EmptyBuildCheck


class CheckRunner:
    def __init__(self, combine):
        self.combine = combine
        self.check_classes = (
            EmptyBuildCheck,
        )
        self.checks = [x(self.combine) for x in self.check_classes]
        self.state = states.UNKNOWN

    @property
    def succeeded(self):
        return self.state == states.SUCCEEDED

    @property
    def failed_checks(self):
        return [x for x in self.checks if x.state == states.FAILED]

    def run(self):
        self.state = states.RUNNING

        for check in self.checks:
            check.run()

        checks_succeeded = all([x.state == states.SUCCEEDED for x in self.checks])
        if checks_succeeded:
            self.state = states.SUCCEEDED
        else:
            self.state = states.FAILED



import os

from .core import Check
from . import states


class EmptyBuildCheck(Check):
    """Ensures that the output directory actually has files in it"""
    def run(self):
        if not os.listdir(self.combine.output_path):
            self.state = states.FAILED
        else:
            self.state = states.SUCCEEDED



import os
import datetime

import yaml


class Config:
    def __init__(self, path):
        self.path = path
        self.data = {}

        if os.path.exists(self.path):
            self.data = yaml.safe_load(open(self.path, 'r'))
            if self.data is None:
                # allow an empty file to start with
                self.data = {}

    def get_variables(self):
        variables = self.default_variables()
        user_variables = self.data.get('variables', {})
        variables.update(user_variables)
        return variables

    def default_variables(self):
        return {
            'now': datetime.datetime.now  # as a function
        }

    def get_commands(self, command_type):
        cmds = self.data.get(command_type, [])
        if not isinstance(cmds, list):
            cmds = [cmds]
        return cmds



import os

from .core import File
from .ignored import IgnoredFile
from .html import HTMLFile
from .sass import SassFile


def file_class_for_path(path):
    _, ext = os.path.splitext(path)

    if os.path.basename(path).startswith('_') or os.path.basename(path).startswith('.'):
        return IgnoredFile

    classes = {
        '.html': HTMLFile,
        '.sass': SassFile,
        '.scss': SassFile,
    }

    if ext in classes:
        return classes[ext]

    return File
